#lang racket/base
(require racket/cmdline)

(define fail (make-continuation-prompt-tag))
(define flip (make-continuation-prompt-tag))

(define-syntax-rule (thunk e)
  (lambda () e))

(define-syntax-rule (shift0-at tag id expr ...)
  (call-with-composable-continuation
   (lambda (k)
     (abort-current-continuation
      tag
      (lambda ()
        (let ([id (lambda vals
                    (call-with-continuation-prompt
                     (lambda ()
                       (apply k vals))
                     tag
                     (lambda (thunk) (thunk))))])
          expr ...))))
   tag))

(define-syntax-rule (prompt0-at prompt expr ...)
  (call-with-continuation-prompt
   (lambda () expr ...)
   prompt
   (lambda (thunk) (thunk))))

(define empty '())

(define (do-fail) (shift0-at fail k empty))
(define (do-flip) (shift0-at flip k (append (k #true) (k #false))))

(define (choice n)
  (if (< n 1)
      (do-fail)
      (if (do-flip)
          (choice (- n 1))
          n)))

(define (triple n s)
  (let* ([i (choice n)]
         [j (choice (- i 1))]
         [k (choice (- j 1))])
    (if (= s (+ i j k))
        (list i j k)
        (do-fail))))


(define (to-list prog)
  (prompt0-at fail (prog)))

(define (collect prog)
  (prompt0-at flip (prog)))

(define (handled-triples-list n s)
  (collect (thunk
            (to-list (thunk
                      (list (triple n s)))))))

(command-line
 #:args (N)
 (let ([N (string->number N)])
   (time (length (handled-triples-list N N)))))
